import asyncio  # flake8: noqa
from typing import Mapping, Optional


Headers = Mapping[str, str]
OptStr = Optional[str]
OptTs = Optional[float]
OptInt = Optional[int]
OptBool = Optional[bool]
OptLoop = Optional[asyncio.AbstractEventLoop]



from typing import Any, Dict, List, NamedTuple, Optional, TypeVar

from .helpers import CONSUMER, PRODUCER, Endpoint, TraceContext, filter_none
from .mypy_types import OptInt, OptStr  # flake8: noqa


Annotation = NamedTuple("Annotation", [("value", str), ("timestamp", int)])


def _endpoint_asdict(endpoint: Endpoint) -> Dict[str, Any]:
    return filter_none(endpoint._asdict())


T = TypeVar("T", bound="Record")


class Record:
    def __init__(self: T, context: TraceContext, local_endpoint: Endpoint) -> None:
        self._context = context
        self._local_endpoint = _endpoint_asdict(local_endpoint)
        self._finished = False

        self._name = "unknown"
        self._kind: OptStr = None
        self._timestamp: OptInt = None
        self._duration: OptInt = None
        self._remote_endpoint: Optional[Dict[str, Any]] = None
        self._annotations: List[Annotation] = []
        self._tags: Dict[str, str] = {}

    @property
    def context(self) -> TraceContext:
        return self._context

    def start(self: T, ts: int) -> T:
        self._timestamp = ts
        return self

    def finish(self: T, ts: OptInt) -> T:
        if self._finished:
            return self
        if self._timestamp is None:
            raise RuntimeError("Record should be started first")
        if ts is not None and self._kind not in (PRODUCER, CONSUMER):
            self._duration = max(ts - self._timestamp, 1)
        self._finished = True
        return self

    def name(self: T, n: str) -> T:
        self._name = n
        return self

    def set_tag(self: T, key: str, value: Any) -> T:
        self._tags[key] = str(value)
        return self

    def annotate(self: T, value: Optional[str], ts: int) -> T:
        self._annotations.append(Annotation(str(value), int(ts)))
        return self

    def kind(self: T, kind: str) -> T:
        self._kind = kind
        return self

    def remote_endpoint(self: T, endpoint: Endpoint) -> T:
        self._remote_endpoint = _endpoint_asdict(endpoint)
        return self

    def asdict(self) -> Dict[str, Any]:
        c = self._context
        rec = {
            "traceId": c.trace_id,
            "name": self._name,
            "parentId": c.parent_id,
            "id": c.span_id,
            "kind": self._kind,
            "timestamp": self._timestamp,
            "duration": self._duration,
            "debug": c.debug,
            "shared": c.shared,
            "localEndpoint": self._local_endpoint,
            "remoteEndpoint": self._remote_endpoint,
            "annotations": [a._asdict() for a in self._annotations],
            "tags": self._tags,
        }
        return filter_none(rec, ["kind"])



from abc import ABCMeta, abstractmethod
from types import TracebackType
from typing import TYPE_CHECKING, List, Optional, Type, TypeVar

from .constants import ERROR
from .helpers import Endpoint, TraceContext, make_timestamp
from .mypy_types import OptInt, OptStr, OptTs
from .record import Record


if TYPE_CHECKING:
    from .tracer import Tracer  # flake8: noqa


T = TypeVar("T", bound="SpanAbc")


class SpanAbc(metaclass=ABCMeta):
    @property
    @abstractmethod
    def is_noop(self: T) -> bool:
        return True  # pragma: no cover

    @property
    @abstractmethod
    def context(self: T) -> TraceContext:
        pass  # pragma: no cover

    @property
    @abstractmethod
    def tracer(self: T) -> "Tracer":
        pass  # pragma: no cover

    @abstractmethod
    def start(self: T, ts: OptTs = None) -> T:
        pass  # pragma: no cover

    @abstractmethod
    def finish(self: T, ts: OptTs = None, exception: Optional[Exception] = None) -> T:
        pass  # pragma: no cover

    @abstractmethod
    def remote_endpoint(
        self: T,
        servce_name: OptStr,
        *,
        ipv4: OptStr = None,
        ipv6: OptStr = None,
        port: OptInt = None
    ) -> T:
        pass  # pragma: no cover

    @abstractmethod
    def tag(self: T, key: str, value: str) -> T:
        pass  # pragma: no cover

    @abstractmethod
    def annotate(self: T, value: Optional[str], ts: OptTs = None) -> T:
        pass  # pragma: no cover

    @abstractmethod
    def kind(self: T, span_kind: str) -> T:
        pass  # pragma: no cover

    @abstractmethod
    def name(self: T, span_name: str) -> T:
        pass  # pragma: no cover

    @abstractmethod
    def new_child(self: T, name: OptStr = None, kind: OptStr = None) -> T:
        pass  # pragma: no cover

    def __enter__(self: T) -> T:
        self.start()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[Exception]],
        exc_value: Optional[Exception],
        traceback: Optional[TracebackType],
    ) -> None:
        self.finish(exception=exc_value)


class NoopSpan(SpanAbc):
    def __init__(
        self,
        tracer: "Tracer",
        context: TraceContext,
        ignored_exceptions: Optional[List[Type[Exception]]] = None,
    ) -> None:
        self._context = context
        self._tracer = tracer
        self._ignored_exceptions = ignored_exceptions or []

    @property
    def is_noop(self) -> bool:
        return True

    @property
    def context(self) -> TraceContext:
        return self._context

    @property
    def tracer(self) -> "Tracer":
        return self._tracer

    def start(self, ts: OptTs = None) -> "NoopSpan":
        return self

    def finish(
        self, ts: OptTs = None, exception: Optional[Exception] = None
    ) -> "NoopSpan":
        return self

    def remote_endpoint(
        self,
        servce_name: OptStr,
        *,
        ipv4: OptStr = None,
        ipv6: OptStr = None,
        port: OptInt = None
    ) -> "NoopSpan":
        return self

    def tag(self, key: str, value: str) -> "NoopSpan":
        return self

    def annotate(self, value: Optional[str], ts: OptTs = None) -> "NoopSpan":
        return self

    def kind(self, span_kind: str) -> "NoopSpan":
        return self

    def name(self, span_name: str) -> "NoopSpan":
        return self

    def new_child(self, name: OptStr = None, kind: OptStr = None) -> "NoopSpan":
        context = self._tracer._next_context(self.context)
        return NoopSpan(self.tracer, context)


class Span(SpanAbc):
    def __init__(
        self,
        tracer: "Tracer",
        context: TraceContext,
        record: Record,
        ignored_exceptions: Optional[List[Type[Exception]]] = None,
    ) -> None:
        self._context = context
        self._tracer = tracer
        self._record = record
        self._ignored_exceptions = ignored_exceptions or []

    @property
    def is_noop(self) -> bool:
        return False

    @property
    def context(self) -> TraceContext:
        return self._context

    @property
    def tracer(self) -> "Tracer":
        return self._tracer

    def start(self, ts: OptTs = None) -> "Span":
        ts = make_timestamp(ts)
        self._record.start(ts)
        return self

    def finish(self, ts: OptTs = None, exception: Optional[Exception] = None) -> "Span":
        if exception is not None:
            if not isinstance(exception, tuple(self._ignored_exceptions)):
                self.tag(ERROR, str(exception))
        ts = make_timestamp(ts)
        self._record.finish(ts)
        self._tracer._send(self._record)
        return self

    def remote_endpoint(
        self,
        servce_name: OptStr,
        *,
        ipv4: OptStr = None,
        ipv6: OptStr = None,
        port: OptInt = None
    ) -> "Span":
        endpoint = Endpoint(servce_name, ipv4, ipv6, port)
        self._record.remote_endpoint(endpoint)
        return self

    def tag(self, key: str, value: str) -> "Span":
        self._record.set_tag(key, value)
        return self

    def annotate(self, value: Optional[str], ts: OptTs = None) -> "Span":
        ts = make_timestamp(ts)
        self._record.annotate(value, ts)
        return self

    def kind(self, span_kind: str) -> "Span":
        self._record.kind(span_kind)
        return self

    def name(self, span_name: str) -> "Span":
        self._record.name(span_name)
        return self

    def new_child(self, name: OptStr = None, kind: OptStr = None) -> "Span":
        span = self.tracer.new_child(self.context)
        if name is not None:
            span.name(name)
        if kind is not None:
            span.kind(kind)
        return span  # type: ignore



# standard tags (binary annotations)
HTTP_HOST = "http.host"
HTTP_METHOD = "http.method"
HTTP_PATH = "http.path"
HTTP_REQUEST_SIZE = "http.request.size"
HTTP_RESPONSE_SIZE = "http.response.size"
HTTP_STATUS_CODE = "http.status_code"
HTTP_URL = "http.url"
HTTP_ROUTE = "http.route"

ERROR = "error"
LOCAL_COMPONENT = "lc"

CLIENT_ADDR = "ca"
MESSAGE_ADDR = "ma"
SERVER_ADDR = "sa"
